/* Copyright (c) 1993, 1994  Washington University in Saint Louis
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met: 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer. 2.
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution. 3. All advertising
 * materials mentioning features or use of this software must display the
 * following acknowledgement: This product includes software developed by the
 * Washington University in Saint Louis and its contributors. 4. Neither the
 * name of the University nor the names of its contributors may be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY WASHINGTON UNIVERSITY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL WASHINGTON
 * UNIVERSITY OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define  HOST_ACCESS  1

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <grp.h>
#include "pathnames.h"
#include "extensions.h"

#ifdef __EMX__
char _PATH_FTPUSERS[256];  /* "/etc/ftpusers" */
char _PATH_FTPACCESS[256]; /* "/usr/local/etc/ftpaccess" */
char _PATH_EXECPATH[256];  /* "/bin/ftp-exec" */
char _PATH_PIDNAMES[256];  /* "/usr/local/daemon/ftpd/ftp.pids-%s" */
char _PATH_CVT[256];       /* "/usr/local/etc/ftpconversions" */
char _PATH_XFERLOG[256];   /* "/usr/adm/xferlog" */
char _PATH_PRIVATE[256];   /* "/etc/ftpgroups" */
char _PATH_UTMP[256];      /* "/etc/utmp" */
char _PATH_WTMP[256];      /* "/usr/adm/wtmp" */
char _PATH_LASTLOG[256];   /* "/usr/adm/lastlog" */
char _PATH_BSHELL[256];    /* "/bin/sh" */
char _PATH_DEVNULL[256];   /* "/dev/null" */
char _PATH_FTPHOSTS[256];  /* "/usr/local/etc/ftphosts" */
char _PATH_SHELLS[256];    /* "/etc/shells" */
#endif

int anonymous = 0;
int guest = 0;
int use_accessfile = 1;
void check_types(char *types);
void check_groups(char *groups);

main()
{
  struct stat  sbuf;
  char        *sp;
  char         buf[1024];

#ifdef __EMX__
    {
    char buf[256], *ptr, *etc, *bin, *tmp, *log;
    etc = getenv("ETC");
    if( !etc ) etc = "c:/tcpip/etc";
    bin = getenv("BIN");
    if( !bin ) bin = "c:/tcpip/wu-ftpd/bin";
    log = getenv("LOG");
    if( !log ) log = "c:/tcpip/wu-ftpd/log";
    tmp = getenv("TMP");
    if( !tmp ) tmp = "c:/tcpip/tmp";
    
#define SET_PATH_VAR(name, strname, root, value)            \
    if( ptr=getenv(strname) ) strncpy(name,ptr,256);        \
    else strcpy(name,root), strcat(name,value);
    
    SET_PATH_VAR(_PATH_FTPUSERS,"_PATH_FTPUSERS",etc,"/ftpusers" )
    SET_PATH_VAR(_PATH_FTPACCESS,"_PATH_FTPACCESS",etc,"/ftpaccess" )
    SET_PATH_VAR(_PATH_EXECPATH,"_PATH_EXECPATH",bin,"" )
    SET_PATH_VAR(_PATH_PIDNAMES,"_PATH_PIDNAMES",log,"/ftp.pids-%s" )
    SET_PATH_VAR(_PATH_CVT,"_PATH_CVT",etc,"/ftpconversions" )
    SET_PATH_VAR(_PATH_XFERLOG,"_PATH_XFERLOG",log,"/xferlog" )
    SET_PATH_VAR(_PATH_PRIVATE,"_PATH_PRIVATE",etc,"/ftpgroups" )
    SET_PATH_VAR(_PATH_UTMP,"_PATH_UTMP",log,"/utmp" )
    SET_PATH_VAR(_PATH_WTMP,"_PATH_WTMP",log,"/wtmp" )
    SET_PATH_VAR(_PATH_LASTLOG,"_PATH_LASTLOG",log,"/lastlog" )
    SET_PATH_VAR(_PATH_BSHELL,"_PATH_BSHELL",bin,"/sh" )
    SET_PATH_VAR(_PATH_DEVNULL,"_PATH_DEVNULL","/dev","/null" )
    SET_PATH_VAR(_PATH_FTPHOSTS,"_PATH_FTPHOSTS",etc,"/ftphosts" )
    SET_PATH_VAR(_PATH_SHELLS,"_PATH_SHELLS",etc,"/shells" )
#undef SET_PATH_VAR
    }
#endif
    
  /* _PATH_FTPUSERS   */
  fprintf(stdout, "Checking _PATH_FTPUSERS :: %s\n", _PATH_FTPUSERS);
  if ( (stat(_PATH_FTPUSERS, &sbuf)) < 0 )
    printf("I can't find it... look in doc/examples for an example.\n");
  else
    printf("ok.\n");

  /* _PATH_FTPACCESS  */
  fprintf(stdout, "\nChecking _PATH_FTPACCESS :: %s\n", _PATH_FTPACCESS);
  if ( (stat(_PATH_FTPACCESS, &sbuf)) < 0 )
    printf("I can't find it... look in doc/examples for an example.\n");
  else {
    struct aclmember *entry;
    struct stat st;

    printf("ok.\n");
    
    if (!readacl(_PATH_FTPACCESS)) {
      printf("Error reading file ftpaccess\n");
      exit(1);
    }
    (void) parseacl();

    entry = NULL;
    if( getaclentry("autogroup",&entry) )
      printf("Warning: autogroup keyword is obsolet for OS/2\n");
    
    entry = NULL;
    while( getaclentry("class",&entry) ) {
      char *p, *s;
      printf("class %s found...", ARG0);
      check_types(ARG1);
      printf("\n");
    }
    
    entry = NULL;
    while( getaclentry("deny",&entry) ) {
      printf("deny %s found\n", ARG0);
      if( stat(ARG1, &sbuf) < 0 ) 
        printf("Error: file %s not found\n", ARG1);
    }
    
    entry = NULL;
    if( getaclentry("guestgroup",&entry) ) {
      int i;
      printf("guestgroup keyword found\n");
      for(i=0; (i < MAXARGS) && ARG[i]; i++)
        if( getgrnam(ARG[i]) )
          printf("Group %s ok\n", ARG[i]);
    }
    
    entry = NULL;
    while( getaclentry("limit",&entry) ) {
      struct aclmember *classentry = 0;
      printf("limit %s found\n", ARG0);
      while( !getaclentry("class",&classentry) )
        if( !strcmp(classentry->arg[0],ARG0) ) break;
      if( !entry ) 
        printf("Error: class %s invalid\n", ARG0);
      if( stat(ARG3, &sbuf) < 0 ) 
        printf("Error: file %s not found\n", ARG3);
    }
    
    entry = NULL;
    if( getaclentry("loginfails",&entry) ) {
      int i;
      printf("loginfails keyword found...");
      if( !sscanf(ARG0,"%d", &i) )
        printf("number error\n");
      else
        printf("%d\n", i);
    }
    
    entry = NULL;
    if( getaclentry("private",&entry) )
      printf("Warning: private keyword is obsolet for OS/2\n");
    
    entry = NULL;
    while( getaclentry("banner",&entry) ) {
      printf("banner %s found\n", ARG0);
      if( stat(ARG0, &sbuf) < 0 ) 
        printf("Error: file %s not found\n", ARG0);
    }
    
    entry = NULL;
    while( getaclentry("email",&entry) ) {
      printf("email %s found\n", ARG0);
    }
    
    entry = NULL;
    while( getaclentry("message",&entry) ) {
      int i;
      printf("message %s %s found\n", ARG0, ARG1);
      if( !stricmp(ARG1,"login") && stat(ARG0, &sbuf) < 0 ) 
        printf("Error: file %s not found\n", ARG0);
      for(i=2; (i < MAXARGS) && ARG[i]; i++) {
        struct aclmember *classentry = 0;
        while( getaclentry("class",&classentry) )
          if( !strcmp(classentry->arg[0],ARG[i]) ) break;
        if( !classentry ) 
          printf("Error: class %s invalid\n", ARG[i]);
      }
    }
    
    entry = NULL;
    while( getaclentry("readme",&entry) ) {
      int i;
      printf("readme %s %s found\n", ARG0, ARG1);
      if( !stricmp(ARG1,"login") && stat(ARG0, &sbuf) < 0 ) 
        printf("Error: file %s not found\n", ARG0);
      for(i=2; (i < MAXARGS) && ARG[i]; i++) {
        struct aclmember *classentry = 0;
        while( getaclentry("class",&classentry) )
          if( !strcmp(classentry->arg[0],ARG[i]) ) break;
        if( !classentry ) 
          printf("Error: class %s invalid\n", ARG[i]);
      }
    }
    
    entry = NULL;
    while( getaclentry("log",&entry) ) {
      char *p, *s;
      printf("log %s found...", ARG0);
      if( !strcmp(ARG0,"commands") ) {
        check_types(ARG1);
        printf("\n");
      }
      else if( !strcmp(ARG0,"transfers") ) {
        check_types(ARG1);
        printf("\t");
        for( p=s=ARG2;*p; p++) {
          while( *p && *p != ',' ) p++;
          if( !strncmp("inbound",s,p-s) ) printf("inbound,");
          else if( !strncmp("outbound",s,p-s) ) printf("outbound,");
          else printf("\nError: direction %s is invalid\n", s);
          s = p+1;
          if( *p==0 ) break;
        }
      }
      else
        printf("Error: subkeyword %s is invalid\n", ARG0);
      printf("\n");
    }
    
    entry = NULL;
    while( getaclentry("alias",&entry) ) {
      printf("alias %s found\n", ARG0);
      if( (stat(ARG1, &sbuf) < 0) && !S_ISDIR(sbuf.st_mode) ) 
        printf("Error: path %s not found\n", ARG1);
    }
    
    entry = NULL;
    while( getaclentry("cdpath",&entry) ) {
      printf("cdpath %s found\n", ARG0);
      if( (stat(ARG0, &sbuf) < 0) && !S_ISDIR(sbuf.st_mode) ) 
        printf("Error: path %s not found\n", ARG0);
    }
    
    entry = NULL;
    while( getaclentry("compress",&entry) ) {
      int i;
      printf("compress %s found\n", ARG0);
      for(i=1; (i < MAXARGS) && ARG[i]; i++) {
        struct aclmember *classentry = 0;
        while( getaclentry("class",&classentry) )
          if( !strcmp(classentry->arg[0],ARG[i]) ) break;
        if( !classentry ) 
          printf("Error: class %s invalid\n", ARG[i]);
      }
    }
    
    entry = NULL;
    while( getaclentry("tar",&entry) ) {
      int i;
      printf("compress %s found\n", ARG0);
      for(i=1; (i < MAXARGS) && ARG[i]; i++) {
        struct aclmember *classentry = 0;
        while( getaclentry("class",&classentry) )
          if( !strcmp(classentry->arg[0],ARG[i]) ) break;
        if( !classentry ) 
          printf("Error: class %s invalid\n", ARG[i]);
      }
    }
    
    entry = NULL;
    while( getaclentry("shutdown",&entry) ) {
      printf("shutdown %s found\n", ARG0);
      if( stat(ARG0, &sbuf) < 0 ) 
        printf("Warning: file %s not found\n", ARG0);
    }
    
    entry = NULL;
    while( getaclentry("chmod",&entry) ) {
      char *p, *s;
      printf("chmod %s found...", ARG0);
      check_types(ARG1);
      printf("\n");
    }
    
    entry = NULL;
    while( getaclentry("delete",&entry) ) {
      char *p, *s;
      printf("delete %s found...", ARG0);
      check_types(ARG1);
      printf("\n");
    }
    
    entry = NULL;
    while( getaclentry("overwrite",&entry) ) {
      char *p, *s;
      printf("overwrite %s found...", ARG0);
      check_types(ARG1);
      printf("\n");
    }
    
    entry = NULL;
    while( getaclentry("rename",&entry) ) {
      char *p, *s;
      printf("rename %s found...", ARG0);
      check_types(ARG1);
      printf("\n");
    }
    
    entry = NULL;
    while( getaclentry("umask",&entry) ) {
      char *p, *s;
      printf("umask %s found...", ARG0);
      check_types(ARG1);
      printf("\n");
    }
    
    entry = NULL;
    while( getaclentry("passwd-check",&entry) ) {
      char *p, *s;
      printf("passwd-check found...");
      if( !strcmp(ARG0,"none") ) printf("none\t");
      else if( !strcmp(ARG0,"trivial") ) printf("trivial\t");
      else if( !strcmp(ARG0,"rfc822") ) printf("rfc822\t");
      else printf("\nError: subkeyword %s is invalid\n", ARG0);
      if( !strcmp(ARG1,"enforce") ) printf("enforce\n");
      else if( !strcmp(ARG1,"warn") ) printf("warn\n");
      else printf("\nError: subkeyword %s is invalid\n", ARG1);
    }
    
    entry = NULL;
    while( getaclentry("path-filter",&entry) ) {
      char *p, *s;
      printf("path-filter found...");
      check_types(ARG0);
      printf("\n");
      if( stat(ARG1, &sbuf) < 0 ) 
        printf("Warning: file %s not found\n", ARG0);
    }
    
    entry = NULL;
    while( getaclentry("upload",&entry) ) {
      char path[MAXPATHLEN+1];
      printf("upload `%s %s %s' found\n", ARG0, ARG1, ARG2);
      if( (stat(ARG0, &sbuf) < 0) && !S_ISDIR(sbuf.st_mode) ) 
        printf("Error: path %s not found\n", ARG0);
      if( *ARG1 == '*' ) continue;
      strcpy(path,ARG0);
      if( path[strlen(path)-1] == '/' || path[strlen(path)-1] == '\\' )
        path[strlen(path)-1] = 0;
      strcat(path,ARG1);
      if( (stat(path, &sbuf) < 0) && !S_ISDIR(sbuf.st_mode) ) 
        printf("Error: path %s not found\n", path);
    }
    
    entry = NULL;
    while( getaclentry("lsshort",&entry) ) {
      printf("lsshort %s found\n", ARG0);
      if( stat(ARG0, &sbuf) < 0 ) 
        printf("Error: file %s not found\n", ARG0);
    }
    
    entry = NULL;
    while( getaclentry("lslong",&entry) ) {
      printf("lslong %s found\n", ARG0);
      if( stat(ARG0, &sbuf) < 0 ) 
        printf("Error: file %s not found\n", ARG0);
    }
    
    entry = NULL;
    while( getaclentry("passwd",&entry) ) {
      int i;
      struct aclmember *group;
      
      printf("passwd found\n", ARG0);
      if( ARG0 ) printf("\tUser %s\n", ARG0);
      else { 
        printf("\nError: username missed\n");
        continue;
      }
      if( ARG1 ) printf("\tpassword *****\n");
      else { 
        printf("\nError: password missed\n");
        continue;
      }
      if( ARG2 && sscanf(ARG2,"%d",&i) ) printf("\tuid %d\n",i);
      else { 
        printf("\nError: uid missed or invalid\n");
        continue;
      }
      if( ARG3 && sscanf(ARG3,"%d",&i) ) printf("\tgid %d",i);
      else { 
        printf("\nError: gid missed or invalid\n");
        continue;
      }
      group = NULL;
      while( getaclentry("group",&group)) {
        int g;
        if( sscanf(group->arg[1],"%d",&g) && i==g ) {
          printf("\t(%s)\n", group->arg[0]);
          break;
        }
      }
      if( !group ) {
        printf("Error: no group with id %d specifyed\n", i);
      }
      if( ARG4 ) printf("\thome dir %s\n", ARG4);
      else { 
        printf("\nError: home dir missed\n");
        continue;
      }
      if( (stat(ARG4, &sbuf) < 0) && !S_ISDIR(sbuf.st_mode) ) 
        printf("\nError: path %s not found\n", ARG0);
      if( ARG5 ) {
        printf("\tshell %s\n", ARG5);
        printf("\tbe shure %s presented in %s\n", ARG5, _PATH_SHELLS);
      }
      if( ARG5 && ((stat(ARG5, &sbuf) < 0)) ) 
        printf("\nError: shell %s not found\n", ARG0);
    }

    entry = NULL;
    while( getaclentry("group",&entry) ) {
      struct group *gr;
      char **gm;
      int i;
      
      printf("group found\n", ARG0);
      if( ARG0 ) printf("\tGroup %s\n", ARG0);
      else { 
        printf("\nError: group name missed\n");
        continue;
      }
      if( ARG1 && sscanf(ARG1,"%d",&i) ) printf("\tgid %d\n",i);
      else { 
        printf("\nError: gid missed or invalid\n");
        continue;
      }
      gr = getgrnam(ARG0);
      printf("\tmembers: ");
      for( gm = gr->gr_mem; gm && *gm && **gm; gm++ ) printf("%s,",*gm);
      printf("\n");
    }
  }
  /* _PATH_PIDNAMES   */
  fprintf(stdout, "\nChecking _PATH_PIDNAMES :: %s\n", _PATH_PIDNAMES);
  strcpy(buf, _PATH_PIDNAMES);
  sp = (char *)strrchr(buf, '/');
  *sp = '\0';
  if ( (stat(buf, &sbuf)) < 0 ) {
    printf("I can't find it...\n");
    printf("You need to make this directory [%s] in order for\n",buf);
    printf("the limit and user count functions to work.\n");
  } else
    printf("ok.\n");

  /* _PATH_CVT        */
  fprintf(stdout, "\nChecking _PATH_CVT :: %s\n", _PATH_CVT);
  if ( (stat(_PATH_CVT, &sbuf)) < 0 ) {
    printf("I can't find it... look in doc/examples for an example.\n");
  } else {
    FILE *cvt;
    char prg[MAXPATHLEN+1];
    
    if( !(cvt=fopen(_PATH_CVT,"rt")) ) {
      printf("Can't open file %s : %s\n", _PATH_CVT, strerror[errno]);
      exit(1);
    }
    while( fgets(prg,MAXPATHLEN,cvt) ) {
      char *s, *p;
      int i;
      
      for(i=0, s=prg; i < 4; i++, s++) 
        while(*s && *s != ':' ) 
          s++;
      if( i != 4 ) {
        printf("Error: format error in ftpconversions file\n");
        exit(1);
      }
      if( s[1] == ':' ) {
        printf("Error: you must use `$' instead of ':' for drive lette delimiting\n");
        exit(1);
      }
      if( s[1] != '$' ) {
        printf("Warning: using programms without drive specified is unsecure\n");
      } 
      else s[1] = ':';
      
      for( p = s; p && *p && !isspace(*p); p++);
      *p = 0;
      if( stat(s, &sbuf) < 0 ) {
        printf("Error: can't find programm %s\n", s);
      }
      else printf("%s - ok.\n",s);
    }
    printf("ok.\n");
  }

  /* _PATH_XFERLOG    */
  fprintf(stdout, "\nChecking _PATH_XFERLOG :: %s\n", _PATH_XFERLOG);
  if ( (stat(_PATH_XFERLOG, &sbuf)) < 0 ) {
    printf("I can't find it... \n");
    printf("Don't worry, it will be created automatically by the\n");
    printf("server if you do transfer logging.\n");
  } else
    printf("ok.\n");

  /* _PATH_PRIVATE    */
  fprintf(stdout, "\nChecking _PATH_PRIVATE :: %s\n", _PATH_PRIVATE);
  if ( (stat(_PATH_PRIVATE, &sbuf)) < 0 ) {
    printf("I can't find it... look in doc/examples for an example.\n");
    printf("You only need this if you want SITE GROUP and SITE GPASS\n");
    printf("functionality. If you do, you will need to edit the example.\n");
  } else
    printf("ok.\n");

  /* _PATH_FTPHOSTS   */
  fprintf(stdout, "\nChecking _PATH_FTPHOSTS :: %s\n", _PATH_FTPHOSTS);
  if ( (stat(_PATH_FTPHOSTS, &sbuf)) < 0 ) {
    printf("I can't find it... look in doc/examples for an example.\n");
    printf("You only need this if you are using the HOST ACCESS features\n");
    printf("of the server.\n");
  } else
    printf("ok.\n");

  /* _PATH_EXECPATH   */
  fprintf(stdout, "\nChecking _PATH_EXECPATH :: %s\n", _PATH_EXECPATH);
  sprintf(buf,"%s/%s", _PATH_EXECPATH, "ls.exe");
  if ( (stat(buf, &sbuf)) < 0 ) {
    printf("I can't find '%s'... It must be present!\n", buf);
    printf("Or you may specify 'lsshort' & 'lslong' in ftpaccess file\n");
  } else
    printf("'%s'ok.\n", buf);
  sprintf(buf,"%s/%s", _PATH_EXECPATH, "so_close.exe");
  if ( (stat(buf, &sbuf)) < 0 ) {
    printf("I can't find '%s'... It should be present, \n\
       or you'll run out of sockets ;)\n", buf);
  } else
    printf("'%s'ok.\n", buf);
  /*
  sprintf(buf,"%s/%s", _PATH_EXECPATH, "index.exe");
  if ( (stat(buf, &sbuf)) < 0 ) {
    printf("I can't find '%s'... It's used for 'SITE INDEX' cmd\n", buf);
  } else
    printf("'%s'ok.\n", buf);
  */
  printf("Also be shure all programs you use with 'SITE EXEC' is\n");
  printf("in '%s' dir\n", _PATH_EXECPATH);

  /* _PATH_UTMP   */
  fprintf(stdout, "\nChecking _PATH_UTMP :: %s\n", _PATH_UTMP);
  if ( (stat(_PATH_UTMP, &sbuf)) < 0 ) {
    printf("I can't find '%s'... Logging to 'utmp' will be disabled\n", _PATH_UTMP);
  } else
    printf("ok.\n");

  /* _PATH_WTMP   */
  fprintf(stdout, "\nChecking _PATH_WTMP :: %s\n", _PATH_WTMP);
  if ( (stat(_PATH_WTMP, &sbuf)) < 0 ) {
    printf("I can't find '%s'... Logging to 'wtmp' will be disabled\n", _PATH_WTMP);
  } else
    printf("ok.\n");

   /* _PATH_LASTLOG   */
  fprintf(stdout, "\nChecking _PATH_LASTLOG :: %s\n", _PATH_LASTLOG);
  if ( (stat(_PATH_LASTLOG, &sbuf)) < 0 ) {
    printf("I can't find '%s'... Logging to 'lastlog' will be disabled\n", _PATH_LASTLOG);
  } else
    printf("ok.\n");

   /* _PATH_BSHELL   */
  fprintf(stdout, "\nChecking _PATH_BSHELL :: %s\n", _PATH_BSHELL);
  if ( (stat(_PATH_BSHELL, &sbuf)) < 0 ) {
    printf("I can't find '%s'... Don't worry, I will use %COMSPEC%\n", _PATH_BSHELL);
  } else
    printf("ok.\n");

   /* _PATH_SHELLS   */
  fprintf(stdout, "\nChecking _PATH_SHELLS :: %s\n", _PATH_SHELLS);
  if ( (stat(_PATH_SHELLS, &sbuf)) < 0 ) {
    printf("I can't find '%s'... Don't worry, I will use %COMSPEC%\n", _PATH_SHELLS);
    printf("But don't set alternative shells in passwd keyword\n");
  } else {
    FILE *shells;
    char shl[MAXPATHLEN+1];
    
    if( !(shells=fopen(_PATH_SHELLS,"rt")) ) {
      printf("Can't open file %s : %s\n", _PATH_SHELLS, strerror[errno]);
      exit(1);
    }
    while( fgets(shl,MAXPATHLEN,shells) ) {
      char *s;
      if( shl[0] == '#' || shl[0] == 0 || isspace(shl[0]) ) continue;
      for( s=shl+strlen(shl)-1; isspace(*s); s--) *s = 0;
      if( stat(shl, &sbuf) < 0 ) {
        printf("Error: can't find shell %s\n", shl);
      }
      else printf("%s - ok.\n",shl);
      if( shl[1] != ':' ) 
        printf("Warning: using shell %s without drive letter is unsecure\n");
    }
    printf("ok.\n");
  }
}


void check_types(char *types) {
  char *p, *s;
  for( p=s=types;*p; p++) {
    while( *p && *p != ',' ) p++;
    if( !strncmp("real",s,p-s) ) printf("real,");
    else if( !strncmp("guest",s,p-s) ) printf("guest,");
    else if( !strncmp("anonymous",s,p-s) ) printf("anonymous,");
    else printf("\nError: type %s is invalid\n", s);
    s = p+1;
    if( *p==0 ) break;
  }
}