/* pwd.c (emx+gcc) -- Copyright (c) 1990-1995 by Eberhard Mattes */

#include <sys/emx.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>

#include "extensions.h"

extern int anonymous, guest;
int os2security = 0xffff;
#define OS2S_SERVSTRIP (4)
#define OS2S_FULLSTRIP (2)
#define OS2S_DRIVESTRIP (1)

static void path_assign (char *dst, const char *src, size_t dst_size)
{
  size_t i;

  i = 0;
  while (*src != 0 && i < dst_size - 1)
    {
      if (*src == '\\')
        dst[i] = '/';
      else
        dst[i] = *src;
      ++i; ++src;
    }
  dst[i] = 0;
}


struct passwd *_getpw (uid_t uid, const char *name)
{
  static struct passwd tmp_pwd;
  static char tmp_name[L_cuserid];
  static char tmp_dir[70];
  static char tmp_shell[70];
  const char *dir;
  const char *shell;
  struct aclmember *entry = NULL;

  if( !name || name[0] == 0 ) {
    while(entry = getaclentry("passwd", &entry)) 
      if( atoi(ARG2) == uid ) { name = ARG0; goto fill_pw; }
  } else {
    while(entry = getaclentry("passwd", &entry)) 
      if( !strcmp(ARG0,name )) { uid = atoi(ARG2); goto fill_pw; }
  }
  return 0;
  
fill_pw:

  dir = ARG4;
  if (dir == NULL)
    dir = "/";

  shell = ARG5;
  if (shell == NULL)
    shell = getenv ("COMSPEC");
  if (shell == NULL)
    shell = "C:/OS2/CMD.EXE";

  _strncpy (tmp_name, name, sizeof (tmp_name));
  path_assign (tmp_dir, dir, sizeof (tmp_dir));
  path_assign (tmp_shell, shell, sizeof (tmp_shell));
  tmp_pwd.pw_name = tmp_name;
  tmp_pwd.pw_passwd = ARG1;
  tmp_pwd.pw_uid = uid;
  tmp_pwd.pw_gid = atoi(ARG3);
  tmp_pwd.pw_age = "";
  tmp_pwd.pw_comment = tmp_name;
  tmp_pwd.pw_gecos = "";
  tmp_pwd.pw_dir = tmp_dir;
  tmp_pwd.pw_shell = tmp_shell;
  syslog(LOG_DEBUG, "PASSWD user %s, uid %d, gid %d, dir %s, shell %s",
         tmp_pwd.pw_name, tmp_pwd.pw_uid, tmp_pwd.pw_gid, 
         tmp_pwd.pw_dir, tmp_pwd.pw_shell);
  return &tmp_pwd;
}


struct passwd *_getpwent (void)
{
  return NULL;                  /* End of /etc/passwd reached */
}


void _setpwent (void)
{
}


void _endpwent (void)
{
}

struct group *getgrnam (__const__ char *gr) {
  struct aclmember *entry = NULL;
  static char tmp_name[9];
  static char *tmp_members[50], tmp_memb_buf[50][9];
  static struct group sgr;
  char *ptr;
  int n = 0, i = 0, tmp_gid = -1;
  
  if( !gr || gr[0] == 0 ) return 0;
  
  bzero(tmp_members,sizeof(tmp_members));
  bzero(tmp_memb_buf,sizeof(tmp_memb_buf));
  strncpy(tmp_name, gr, sizeof(tmp_name));
  
  /* Get gid and members from group source */
  while(entry = getaclentry("group", &entry))
    if( !strcmp(gr,ARG0) ) break;
  if( !entry ) goto done;
  tmp_gid = atoi(ARG1);
  for(ptr=ARG2; ptr && *ptr; ptr++) {
    tmp_members[n] = tmp_memb_buf[n];
    if( *ptr != ',' && i < 8 )
      tmp_memb_buf[n][i++] = *ptr;
    else {
      tmp_memb_buf[n][i] = 0;
      i = 0;
      n++;
      while( *ptr && ptr != ',' ) ptr++;
    }
  }
    
  /* Append to group all users from passwd source */
  while(entry = getaclentry("passwd", &entry)) {
    if( atoi(ARG3) == tmp_gid ) {
      strncpy(tmp_memb_buf[++n],ARG0,9);
      tmp_members[n] = tmp_memb_buf[n];
    }
  }
done:
  sgr.gr_name = tmp_name;
  sgr.gr_gid = tmp_gid;
  sgr.gr_mem = tmp_members;
  syslog(LOG_DEBUG, "GRNAM group %s, gid %d, members %s %s %s %s %s %s %s %s",
         sgr.gr_name, sgr.gr_gid,
         sgr.gr_mem[0]?sgr.gr_mem[0]:"",
         sgr.gr_mem[1]?sgr.gr_mem[1]:"",
         sgr.gr_mem[2]?sgr.gr_mem[2]:"",
         sgr.gr_mem[3]?sgr.gr_mem[3]:"",
         sgr.gr_mem[4]?sgr.gr_mem[4]:"",
         sgr.gr_mem[5]?sgr.gr_mem[5]:"",
         sgr.gr_mem[6]?sgr.gr_mem[6]:"",
         sgr.gr_mem[7]?sgr.gr_mem[7]:""
         );
  return &sgr;
}


int chroot( char *path ) {
  static char rootdir[256];
  int success = -1;
  success = _chdir2(path);
  if( !getcwd(rootdir, sizeof(rootdir)-1) ) {
    success = -1;
    syslog(LOG_ERR, "cannot chroot to %s: %s", path, strerror(errno));
  }
  return success;
}

char *stripdrive(char *s) {
  char *n;
  
  if( !(anonymous || guest) ) return s;
  
  for(;;) {
    if( strlen(s) < 2 ) return s;
    if( (os2security & OS2S_DRIVESTRIP) && 
             ( s[1] == ':' && isalpha(*s) )
           ) {
      s += 2;
      continue;
    }
    else if( (os2security & OS2S_FULLSTRIP) && (n=strchr(s,':')) ) {
      s = n+1;
      continue;
    }
    else if( (os2security & OS2S_SERVSTRIP) && 
             ((s[0]=='/' || s[0]=='\\') && (s[1]=='/' || s[1]=='\\')) 
           ) {
      if( n = strchr(s+2, '/') ) {
        s = n+1;
        continue;
      }
      else if( n = strchr(s+2, '\\') ) {
        s = n+1;
        continue;
      }
      else s += strlen(s);
    }
    else break;
  }
  return s;
}